@@ -15,7 +15,8 @@ def new(obj, path, value, separator="/"):
     characters in it, they will become part of the resulting
     keys
     """
-    return dpath.path.set(obj, path.lstrip(separator).split(separator), value, create_missing=True)
+    pathobj = dpath.path.path_types(obj, path.lstrip(separator).split(separator))
+    return dpath.path.set(obj, pathobj, value, create_missing=True)
 
 def delete(obj, glob, separator="/", filter=None):
     """
@@ -26,8 +27,7 @@ def delete(obj, glob, separator="/", filter=None):
     """
     deleted = 0
     paths = []
-    print separator
-    for path in dpath.path.search(obj, glob.lstrip(separator).split(separator)):
+    for path in _inner_search(obj, glob.lstrip(separator).split(separator)):
         # These are yielded back, don't mess up the dict.
         paths.append(path)
 
@@ -37,13 +37,13 @@ def delete(obj, glob, separator="/", filter=None):
         for item in path:
             prev = cur
             try:
-                cur = cur[item]
+                cur = cur[item[0]]
             except AttributeError, e:
                 # This only happens when we delete X/Y and the next
                 # item in the paths is X/Y/Z
                 pass
-        if (not filter) or (filter and filter(prev[item])):
-            prev.pop(item)
+        if (not filter) or (filter and filter(prev[item[0]])):
+            prev.pop(item[0])
         deleted += 1
     if not deleted:
         raise dpath.exceptions.PathNotFound("Could not find {} to delete it".format(glob))
@@ -55,7 +55,7 @@ def set(obj, glob, value, separator="/", filter=None):
     to the given value. Returns the number of elements changed.
     """
     changed = 0
-    for path in dpath.path.search(obj, glob.lstrip(separator).split(separator)):
+    for path in _inner_search(obj, glob.lstrip(separator).split(separator)):
         changed += 1
         dpath.path.set(obj, path, value, create_missing=False, filter=filter)
     return changed
@@ -72,22 +72,34 @@ def search(obj, glob, yielded=False, separator="/", filter=None):
 
     def _search_view(obj, glob):
         view = {}
-        for path in dpath.path.search(obj, glob.lstrip(separator).split(separator), dirs=False):
-            val = dpath.path.get(obj, path, view=False)
-            if (not filter) or (filter(val)):
-                dpath.path.set(view, path, val, create_missing=True)
+        for path in _inner_search(obj, glob.lstrip(separator).split(separator), dirs=(filter == None)):
+            try:
+                val = dpath.path.get(obj, path, filter=filter, view=True)
+                merge(view, val)
+            except dpath.exceptions.FilteredValue:
+                print "FilteredValue on %s" % str(path)
+                pass
         return view
 
     def _search_yielded(obj, glob):
-        for path in dpath.path.search(obj, glob.lstrip(separator).split(separator)):
-            val = dpath.path.get(obj, path, view=False)
-            if (not filter) or (filter and filter(val)):
-                yield (separator.join(path), val)
+        for path in _inner_search(obj, glob.lstrip(separator).split(separator), dirs=False):
+            try:
+                val = dpath.path.get(obj, path, view=False, filter=filter)
+                yield (separator.join(map(str, dpath.path.paths_only(path))), val)
+            except dpath.exceptions.FilteredValue:
+                print "FilteredValue on %s" % str(path)
+                pass
 
     if yielded:
         return _search_yielded(obj, glob)
     return _search_view(obj, glob)
 
+def _inner_search(obj, glob, dirs=True, leaves=False):
+    """Search the object paths that match the glob."""
+    for path in dpath.path.paths(obj, dirs, leaves, skip=True):
+        if dpath.path.match(path, glob):
+            yield path
+
 def merge(dst, src, separator="/", filter=None, flags=MERGE_ADDITIVE, _path=""):
     """Merge source into destination. Like dict.update() but performs
     deep merging.
@@ -104,47 +116,42 @@ def merge(dst, src, separator="/", filter=None, flags=MERGE_ADDITIVE, _path=""):
           replaces the destination in this situation.
     """
 
+    if filter:
+        # Having merge do its own filtering is dumb, let search do the
+        # heavy lifting for us.
+        src = search(src, '**', filter=filter)
+        return merge(dst, src)
+
     def _check_typesafe(obj1, obj2, key, path):
         if ( (flags & MERGE_TYPESAFE) and (type(obj1[key]) != type(obj2[key]))):
             raise TypeError("Cannot merge objects of type {} and {} at {}"
                             "".format(type(obj1[key]), type(obj2[key]), path))
 
-
-    def _filter_assign(obj, key, value):
-        if (not filter) or (filter and filter(value)):
-            obj[key] = value
-
     if isinstance(src, dict):
         for (i, v) in enumerate(src):
             _check_typesafe(dst, src, v, separator.join([_path, str(v)]))
 
             if not v in dst:
-                if not isinstance(src[v], (dict, list)):
-                    _filter_assign(dst, v, src[v])
-                else:
-                    dst[v] = src[v]
+                dst[v] = src[v]
             else:
                 if not isinstance(src[v], (dict, list)):
-                    _filter_assign(dst, v, src[v])
+                    dst[v] = src[v]
                 else:
                     merge(dst[v], src[v], filter=filter, flags=flags,
                           _path=separator.join([_path, str(v)]), separator=separator)
     elif isinstance(src, list):
         for (i, v) in enumerate(src):
             _check_typesafe(dst, src, i, separator.join([_path, str(i)]))
-
             dsti = i
             if ( flags & MERGE_ADDITIVE):
                 dsti += len(src)
             if dsti >= len(dst):
                 dst += [None] * (dsti - len(dst) + 1)
             if dst[dsti] == None:
-                _filter_assign(dst, dsti, src[i])
+                dst[dsti] = src[i]
             else:
                 if not isinstance(src[i], (dict, list)):
-                    _filter_assign(dst, dsti, src[i])
+                    dst[dsti] = src[i]
                 else:
                     merge(dst[i], src[i], filter=filter, flags=flags,
                           _path=separator.join(_path, v), separator=separator)
-
-
